package eu.unicore.gateway;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URL;

import javax.servlet.http.HttpServletRequest;

import junit.framework.TestCase;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import eu.unicore.bugsreporter.annotation.FunctionalTest;
import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.util.ProtocolPluginHandler;
import eu.unicore.gateway.util.UpdateProtocolMethod;

public class TestServer extends TestCase {
	static Gateway gw;
	static boolean haveInit=false;
	
	@Override
	protected void setUp() throws Exception{
		File gp = new File("src/test/resources/gateway.properties");
		File sp = new File("src/test/resources/security.properties");
		File cp = new File("src/test/resources/connection.properties");
		gw = new Gateway(gp, cp, sp);
		gw.startGateway();
	}
	
	protected void tearDown()throws Exception{
		Thread.sleep(2000);
		gw.stopGateway();
	}
	public void testSwitchProtocols()throws Exception{
		URL url=new URL("http://localhost:64433/foo");
		HttpClient hc = gw.getClientFactory().makeHttpClient(url);
		UpdateProtocolMethod post=new UpdateProtocolMethod(new ProtocolPluginHandler(){
			public void handle(InputStream is, OutputStream os)throws IOException{
				BufferedReader br=new BufferedReader(new InputStreamReader(is));
				byte[] data="Testdata!!\n".getBytes();
				for(int i=0;i<2;i++){
					System.out.print("send : "+new String(data));
					os.write(data);
					os.flush();
					String line=br.readLine();
					if(line!=null){
						System.out.println("recv : "+line);
						assertTrue(line.contains("Testdata"));
					}
				}
				is.close();
				os.close();
			}
			public String getProtocol(){return "ECHO";}
			public void configure(HttpServletRequest req){}
		});
		
		hc.executeMethod(post);
		assertTrue("Error occured, server replied: "+post.getStatusLine(),post.getStatusCode()==101);
		post.releaseConnection();
	}
	
	@FunctionalTest(id="gw_dynReg", description="Tests whether dynamic registration is possible and if can be disabled.")
	public void testDynamicRegistration()throws Exception{
		gw.getProperties().setProperty(GatewayProperties.KEY_REG_ENABLED, "false");
		int status=doRegister("test","http://localhost:12345");
		assertEquals(HttpStatus.SC_FORBIDDEN,status);
		gw.getProperties().setProperty(GatewayProperties.KEY_REG_ENABLED, "true");
		status=doRegister("test","http://localhost:12345");
		assertEquals(HttpStatus.SC_CREATED,status);
	}
	
	public void testPost()throws Exception{
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBody(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
			assertTrue(post.getResponseHeader("Content-type").getValue().contains("application/xml"));
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		int expectedCode=503;
		
		s1.setStatusCode(expectedCode);
		
		post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBody(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
		    //TODO what is the expected behaviour? Currently the GW always returns 200 OK even
			//if the vsite gives an error code 
			//assertEquals(expectedCode, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		
		s1.stop();
	}
	
	public void testPostNoSOAP()throws Exception{
		Logger.getLogger("unicore.gateway.Servlet").setLevel(Level.DEBUG);
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBodyNoSOAP(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
			assertEquals(HttpStatus.SC_INTERNAL_SERVER_ERROR, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		s1.stop();
	}

	public void testGet()throws Exception{
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		GetMethod get=new GetMethod(url);
		try{
			hc.executeMethod(get);
			System.out.println(get.getStatusLine());
			status=get.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		int expectedCode=503;
		s1.setStatusCode(expectedCode);
		
		get=new GetMethod(url);
		try{
			hc.executeMethod(get);
			System.out.println(get.getStatusLine());
			status=get.getStatusCode();
			assertEquals(expectedCode, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		
		s1.stop();
	}
	/**
	 * get a minimalistic SOAP message
	 * @param to - WS Addressing To header
	 */
	private byte[] getBody(String to)throws IOException{
		ByteArrayOutputStream bos=new ByteArrayOutputStream();
		OutputStreamWriter wr=new OutputStreamWriter(bos);
		
		wr.write("<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\"\n"+
            "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" \n"+
            "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">"+
           "<soap:Header>");
		wr.write("<wsa:To xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">");
		wr.write(to);
		wr.write("</wsa:To>\n");
		wr.write("</soap:Header>");
		wr.write("<soap:Body>test</soap:Body>");
		wr.write("</soap:Envelope>");
		wr.flush();
		return bos.toByteArray();
	}
	
	/**
	 * get non-SOAP body for POSTing
	 * @param to - WS Addressing To header
	 */
	private byte[] getBodyNoSOAP(String to)throws IOException{
		ByteArrayOutputStream bos=new ByteArrayOutputStream();
		OutputStreamWriter wr=new OutputStreamWriter(bos);
		wr.write("P<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\"\n"+
	            "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" \n"+
	            "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">");
		wr.write("PPP");
		wr.flush();
		return bos.toByteArray();
	}
	
	
	
	private int doRegister(String name, String address)throws Exception{
		String url="http://localhost:64433/VSITE_REGISTRATION_REQUEST";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		try{
			post.addParameter("name", name);
			post.addParameter("address", address);
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			return post.getStatusCode();
		}
		finally{
			post.releaseConnection();
		}
	}
	
}
